/* 
   File: multicast.c
   Written By: Rahul Srivastava
   Date: Nov 28, 1994

   Synopsis: this file contains code of the functions to add multicasting
             features to a process
*/

#include <unistd.h>
#include <net/if.h>
#include "multicast.h"

void checkMulticast(int sock);

int debug;
        
mcast_info mcast;

void initMulticast(int sock, int sessionID, int port, struct sockaddr_in *addr)
{
  u_char ttl;
  u_char loopback = 0;

  checkMulticast(sock);

  ttl = (u_char)128;

  map_mcast_address(&mcast, sessionID);
  mcast.port = port;

  addr->sin_family = AF_INET;
  addr->sin_port = port;
  addr->sin_addr.s_addr = inet_addr(mcast.IPaddress);
  
  reusePort(sock);

  setTTLvalue(sock,ttl);
  
  setLoopback(sock, loopback);

  joinGroup(sock, mcast.IPaddress);
}

void closeMulticast(int sock)
{
  leaveGroup(sock, mcast.IPaddress);
}

/*
   Function : leaveGroup

   Input : socket descriptor of process that listens to the group messages
           ID of group to leave

   Synopsis : this function removes the process from the group 

*/

void leaveGroup(int recvSock,char *group)
{
  struct sockaddr_in groupStruct;
  struct ip_mreq dreq;  /* multicast group info structure */

  if((groupStruct.sin_addr.s_addr = inet_addr(group))== -1)
     printf("error in inet_addr\n");

  dreq.imr_multiaddr = groupStruct.sin_addr;
  dreq.imr_interface.s_addr = INADDR_ANY;
  
  if(setsockopt(recvSock,IPPROTO_IP,IP_DROP_MEMBERSHIP,&dreq,sizeof(dreq)) == -1)
  {
    printf("error in leaving group \n");
    exit(-1);
  }
}


/*
   Function : reusePort

   Input  : socket descriptor

   Output : void

   Synopsis : this function sets a socket option that allows multipule processes
              to bind to the same port
*/

void reusePort(int s)
{
  int one=1;
  
  if(setsockopt(s,SOL_SOCKET,SO_REUSEPORT,&one,sizeof(one)) == -1)
  {
    printf("error in setsockopt,SO_REUSEPORT \n");
    exit(-1);
  }
}


/*
   Function : setTTLvalue

   Input :
           socket descriptor
	   address of ttl value to be set
*/

void setTTLvalue(int s,u_char ttl_value)
{
  if(setsockopt(s,IPPROTO_IP,IP_MULTICAST_TTL, &ttl_value,sizeof(u_char)) == -1)
  {
    printf("error in setting ttl value\n");
  }
}


/*
   Function : setLoopback

   Input :
           socket descriptor
	   address of the "loop" variable

	   loop = 1 /* means enable loopback 
	   loop = 0 /* means disable loopback 
		       
   Synopsis : By default, messages sent to the multicast group are looped 
              back to the local host. this function disables that.
	      
   NOTE : by default, loopback is enabled
*/
       
void setLoopback(int s,u_char loop)
{
  if(setsockopt(s,IPPROTO_IP,IP_MULTICAST_LOOP,&loop,sizeof(u_char)) == -1)
  {
    printf("error in disabling loopback\n");
  }
}


/* 
   Function : joinGroup

   Input :
          socket descriptor of process that want's to join the group
	  ID of the group to join

   Synopsis : this function sets the socket option to make the local host 
              join the mulicast group 
*/

void joinGroup(int s, char *group)
{
  struct sockaddr_in groupStruct;
  struct ip_mreq mreq;  /* multicast group info structure */

  /*printf("App: Joined multicast group %s, socket %d\n", group, s);*/
  if((groupStruct.sin_addr.s_addr = inet_addr(group))== -1)
     perror("inet_addr");

  /* check if group address is indeed a Class D address */
  mreq.imr_multiaddr = groupStruct.sin_addr;
  mreq.imr_interface.s_addr = INADDR_ANY;
  
  if(setsockopt(s,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)) == -1)
  {
    perror("joinGroup");
    exit(-1);
  }
}


   
/* 
   Function : map_mcast_address(mcast_info *session, int session_numb)

   Synopsis : this function maps the session number to the multicast address being
              used for the class. 

   modified : By Bjoern J Kvande 4/29/95.
              corrected conversion from int session_numb to string

*/

void map_mcast_address(mcast_info *session, int session_numb)
{
  char s[20];
  
  bzero(session->IPaddress,64);
  strcpy(session->IPaddress,MULTICAST_ADDRESS_BASE);
  sprintf(s,"%i", session_numb);
  strcat(session->IPaddress, s);  
  /*  session->port = MULTICAST_PORT_BASE + session_numb;*/
  session->ttl = DEFAULT_MULTICAST_TTL_VALUE;
  
  if (debug)
     fprintf(stderr,"process joining multicast group %s:%d \n",session->IPaddress,
             session->port);
}


void checkMulticast(int sock)
{
  int n;
  struct ifconf ifc;
  char buf[BUFSIZ]; 
  struct ifreq *ifr;
  struct sockaddr dst;

  ifc.ifc_len = sizeof (buf);
  ifc.ifc_buf = buf;
  if (ioctl(sock, SIOCGIFCONF, (char *) &ifc) < 0) {
    printf("Conf retrieve error\n");
  }  
  
  ifr = ifc.ifc_req; 
  for (n = ifc.ifc_len / sizeof (struct ifreq); --n >= 0;      ifr++) 
    {    
      if (ifr->ifr_addr.sa_family != AF_INET)         
	continue;     
      if (ioctl(sock, SIOCGIFFLAGS, (char *) ifr) < 0) 
	{
	  printf("Flags error\n");
	}
    }
      if (ifr->ifr_flags & IFF_MULTICAST)
	printf("Supporting multicast\n");
      else
	printf("Not supporting multicast\n");
}
